# LightBulb
LightBulb is an open source python framework for auditing web applications firewalls.

## Synopsis

The framework consists of two main algorithms:

* **SFALearner**: An active learning algorithm that infers symbolic representations of automata in the standard membership/equivalence query model.
 
    Active learning algorithms permits the analysis of filter and sanitizer programs remotely, i.e. given only the ability to query the targeted program and observe the output.

* **SFADiff**: A black-box differential testing algorithm based on Symbolic Finite Automata (SFA) learning

    Finding differences between programs with similar functionality is an important security problem as such differences can be used for fingerprinting or creating evasion attacks against security software like Web Application Firewalls (WAFs) which are designed to detect malicious inputs to web applications.

## Motivation

Web Applications Firewalls (WAFs) are fundamental building blocks of modern application security. For example, the PCI standard for organizations handling credit card transactions dictates that any application facing the internet should be either protected by a WAF or successfully pass a code review process. Nevertheless, despite their popularity and importance, auditing web application firewalls remains a challenging and complex task. Finding attacks that bypass the firewall usually requires expert domain knowledge for a specific vulnerability class. Thus, penetration testers not armed with this knowledge are left with publicly available lists of attack strings, like the XSS Cheat Sheet, which are usually insufficient for thoroughly evaluating the security of a WAF product.


## Commands Usage

Main interface commands:
 
 Command       | Description                           
 ------------- | ------------------------------------- 
 core          | Shows available core modules          
 info  \<module\>  | Prints module information             
 library       | Enters library                       
 modules       | Shows available application modules  
 use \<module\>    | Enters module                         
 help          | Prints help                           
 complete      | Prints bash completion command        

Module commands:
 
 Command       | Description                           
 ------------- | ------------------------------------- 
 back          | Go back to main menu         
 info          | Prints  current module information             
 library       | Enters library                       
 options       | Shows available options
 define \<option\>  \<value\>   | Set an option value
 start         | Initiate algoritm   
 complete      | Prints bash completion command    

Library commands:

 Command       | Description                           
 ------------- | ------------------------------------- 
 back          | Go back to main menu     
 info \<folder\\module\>  | Prints requested module information (folder must be located in lightbulb/data/)
 cat \<folder\\module\>  | Prints requested module  (folder must be located in lightbulb/data/)
 modules  \<folder\>     | Shows available library modules in the requested folder (folder must be located in lightbulb/data/)
 search  \<keywords\>    | Searches available library modules using comma separated keywords
 complete      | Prints bash completion command    

## Installation

In order to use the application without complete package installation:

```bash
python setup.py build
pip install -r requirements.txt
python bin/lightbulb
```

In order to perform complete package installation (Only available after registering automata and fsalearn packages to pip repository):

```bash
python setup.py install
lightbulb
```

You can also install it from pip repository (Only available after registering automata. fsalearn and lightbulb packages to pip repository):

```bash
pip install lightbulb
lightbulb
```


## Core Modules Design

A core module complies with the following template:

```python

META = {
        'author': 'Module Author',
        'description': 'Module Description',
        'comments': ['Sample comment']
    }
    
class MODULE_UNIQUE_NAME():

    def __init__(self, params):
        ...
        
    def getresult(self):
        """
        Returns the resulting string
        Args:
           None
        Returns:
           str: The resulting string
        """
    
    def learn(self):
        """Performs algorithm"""
        
    def stats(self):
        """Prints execution statistics"""
        return [('NAME', VALUE)]

    
```


## Application Modules Design

For application modules the minimum functions that is required to be declared:

```python
from lightbulb.core.base import CORE, options_as_dictionary

META = {
        'author': 'Module Author',
        'description': 'Module Description',
        'type':'Algorithm type',
        'options': [
            ('Option name', value, (BOOL) REQUIRED, 'Option description'),
        ],
        'comments': ['Sample comment']
    }
    
class Module(CORE['COREMODULE']):

    def __init__(self, configuration):
        self.module_config = options_as_dictionary(configuration)
        super(Module, self).__init__(self.module_config['Option name'])
    
    def learn(self):
        ...
     
    def stats(self):
        return  [("NAME",VALUE)]

    def getresult(self):
        return  "NAME", VALUE
```

The framework also supports execution in parallel for two core modules. Then. the "Module_A" and "Module_B" classes should be declared instead of the "Module" class:

```python

class Module_A(CORE['COREMODULE']):

    def __init__(self, configuration, channel, cross):
        self.module_config = options_as_dictionary(configuration)
        super(Module, self).__init__(self.module_config['Option name'])
    
    def learn(self):
        ...
     
    def stats(self):
        return  [("NAME",VALUE)]

    def getresult(self):
        return  "NAME", VALUE
        
class Module_B(CORE['COREMODULE']):

    def __init__(self, configuration, channel, cross):
        self.module_config = options_as_dictionary(configuration)
        super(Module, self).__init__(self.module_config['Option name'])
    
    def learn(self):
        ...
     
    def stats(self):
        return  [("NAME",VALUE)]

    def getresult(self):
        return  "NAME", VALUE
```


## Using SFALearner and SFADiff Core Modules 

Currently, the framework supports two core modules, the **'SFALearner'** and the **'SFADiff'**. The the **'SFADiff'** works exaclty as the **'SFALearner'** core module, but allows two parallel instances to communicate and exchange their learned model during the equivalence query functionality. These core modules already define learn(), stats(), and getresult() functions. However, they declare the _query() function that is required to be overriden.

For **SFALearner** the "Module" class should be used and the minimum functions that are required to be overridden are the following:

```python
...
class Module(CORE['SFALearner']):

    def __init__(self, configuration):
        ...
        super(Module, self).__init__(alphabet, file, file_type,  preload)
    
    def _query(self, string):
        """
        Performs a membership query 
        Args:
            string: The string to be tested
        Returns:
            bool: A boolean value indicating that the target parser accepts or rejects the input
        """
        return VALUE
```

For **SFADiff** the "Module_A" and "Module_B" classes should be used and the minimum functions that are required to be overridden are the following:

```python
...
class Module_A(CORE['SFADiff']):

    def __init__(self, configuration, channel, cross):
        ...
        super(Module_A, self).__init__(alphabet, file, file_type, channel, cross, classic,left, right, preload)
    
    def _query(self, string):
        """
        Performs a membership query 
        Args:
            string: The string to be tested
        Returns:
            bool: A boolean value indicating that the target parser accepts or rejects the input
        """
        return VALUE
        
class Module_B(CORE['SFADiff']):

    def __init__(self, configuration, channel, cross):
        ...
        super(Module_B, self).__init__(alphabet, file, file_type, channel, cross, classic,left, right, preload)
    
    def _query(self, string):
        """
        Performs a membership query 
        Args:
            string: The string to be tested
        Returns:
            bool: A boolean value indicating that the target parser accepts or rejects the input
        """
        return VALUE

```


## API Reference

The framework offers a number of built-in query handlers for enhancing the membership queries:

* The **HTTP Handler**:
 
    To initialize the handler:
     
    ```python
    from lightbulb.core.utils.http_handler import HTTPHandler as _HTTPHandler
    httphandler = _HTTPHandler(URL, REQUEST_TYPE, PARAM, BLOCK, BYPASS, PROXY_SCHEME, PROXY_HOST, PROXY_PORT, PROXY_USERNAME, PROXY_PASSWORD, USER_AGENT, REFERER)
    ```
     
    To perform a query:
     
    ```python
    httphandler.do_query(string)
    ```   

* The **SQL Handler**:

    To initialize the handler:
     
    ```python
    from lightbulb.core.utils.sql_handler import SQLHandler as _SQLHandler
    sqlserver = _SQLHandler(HOST, PORT, USERNAME, PASSWORD, DATABASE, PREFIX_QUERY, SQLPARSES)
    ```
     
    To perform a query:
     
    ```python
    sqlserver.do_query(string)
    ```   

* The **Browser Handler**:

    To initialize the handler:
 
    ```python
    from lightbulb.core.utils.browsersocket_hander import BrowserSocketHandler as _BrowserSocketHandler
    browser_socket_handler = _BrowserSocketHandler(WSPORT, WBPORT, BROWSERPARSES)
    ```
     
    To perform a query:
     
    ```python
    browser_socket_handler.do_query(string)
    ```   

* The **Options Parser**:

    ```python
    from lightbulb.core.base import options_as_dictionary
    ```
     
    To perform a query:
     
    ```python
    module_config = options_as_dictionary(configuration)
    ```   

* The **Alphabet**:
    
    ```python
    from lightbulb.core.base import createalphabet
    ```
     
    To retrieve alphabet:
     
    ```python
    createalphabet()
    ```   
    

## Library Modules

A library module consists of two files:
* A file (ending in .y or .json) that contains the actual grammar, regular expression or json configuration. It can also be a folder.
* A file (ending in .py) having the same name, that describes the previous file

The file (ending in .py) must have the following form:

```python
META = {
    'author': 'Module Author',
    'description': 'File/Folder containing ...',
    'type':'Regex/Grammar/Folder/Configuration',
    'comments': []
}
```   
    
  
    
## Tests

Describe and show how to run the tests with code examples.

## Contributors

* George Argyros
* Ioannis Stais

## License

A short snippet describing the license (MIT, Apache, etc.)