"""This module contains all the functions that perform an operation for the CLI"""
import logging
import sys
import os
import imp
from cliff.command import Command
from cliff.lister import Lister
from lightbulb.core.operate import manage
from lightbulb.core.base import options_as_dictionary
sys.path.insert(1, imp.find_module('lightbulb')[1]+ '/modules')

CURRENT = None
CONFIG = {}
META = None

def update_in_alist(alist, key, value):
    """Updates a value in a list of tuples"""
    return [(k, v1, v2, v3) if (k != key) else (key, value, v2, v3) for (k, v1, v2, v3) in alist]

def update_in_alist_inplace(alist, key, value):
    """Updates a list of tuples in place"""
    alist[:] = update_in_alist(alist, key, value)

def check_parameters(config):
    """
    Checks if all required parameters have been set
    Args:
        meta (list): A list of option tuples
    Returns:
        list: A list of option tuples that have not been set
    """
    required = []
    for option in META:
        if option[2] is True and (option[0] not in CONFIG or CONFIG[option[0]] is None):
            required.append((option[0], option[1], option[2]))
    return required

class Use(Command):
    "Entes in an application or core module view"

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """
        Argument parser
        Args:
            prog_name (object): The current program
        Returns:
            object: A parser for the provided arguments
        """
        parser = super(Use, self).get_parser(prog_name)
        parser.add_argument('module', nargs='?', default='.')
        return parser

    def take_action(self, parsed_args):
        """
        Enters in an application or core module view
        Args:
            parsed_args.module (str): The requested module
        Returns:
            None
        """
        global CURRENT, CONFIG, META
        module_name = parsed_args.module
        try:
            META = __import__(module_name).META['options']
            self.app.stdout.write('Entering module ' + module_name + '\n')
        except:
            self.app.stdout.write('Module ' + module_name + ' was not found\n')
            return False
        CURRENT = module_name
        configuration = options_as_dictionary(META)
        for option in configuration:
            if option not in CONFIG:
                CONFIG[option] = configuration[option]
        print "\n"
        return True

class Library(Command):
    "Entes in library view"

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """
        Argument parser
        Args:
            prog_name (object): The current program
        Returns:
            object: A parser for the provided arguments
        """
        parser = super(Library, self).get_parser(prog_name)
        parser.add_argument('module', nargs='?', default='.')
        return parser

    def take_action(self, parsed_args):
        """
        Enters in library view
        """
        pass

class Define(Command):
    "Sets an option value"

    log = logging.getLogger(__name__)


    def get_parser(self, prog_name):
        """
        Argument parser
        Args:
            prog_name (object): The current program
        Returns:
            object: A parser for the provided arguments
        """
        parser = super(Define, self).get_parser(prog_name)
        parser.add_argument('set', nargs='*', default='.')
        return parser

    def take_action(self, parsed_args):
        global CONFIG
        """
        Sets an option value
        Args:
            parsed_args.set[0] (str): The requested option
            parsed_args.set[1] (str): The new value
        Returns:
            None
        """
        CONFIG[parsed_args.set[0]] =  (' '.join(parsed_args.set[1:])).strip()

        print "\n"


class Back(Command):
    "Moves back to main menu view"

    log = logging.getLogger(__name__)


    def take_action(self, parsed_args):
        """
        Moves back to main menu view
        """
        global  CURRENT, META
        print 'Back to main menu '
        CURRENT = None
        META = None
        print "\n"

class LibraryBack(Command):
    "Moves back to previous view"

    log = logging.getLogger(__name__)


    def take_action(self, parsed_args):
        """
        Moves back to previous view
        """
        global  CURRENT, META
        if CURRENT:
            print 'Back to module '+CURRENT
            print "\n"
        else:
            print 'Back to main menu '
            CURRENT = None
            META = None
            print "\n"

class Info(Lister):
    """
    Returns information about currently used module
    """
    log = logging.getLogger(__name__)

    def take_action(self, parsed_args):
        """
        Returns information about the currently used module
        Args:
            parsed_args.module (str): The requested module
        Returns:
            tuple: Details about the module
        """
        module_name = CURRENT
        self.app.stdout.write('Module '+module_name+' Information:\n')
        meta = __import__(module_name).META

        return (('Name', 'Value'),
                ([('Author', meta['author']), ('Description', meta['description']),
                  ('Comments', ', '.join(meta['comments']))]))

class Start(Lister):
    """Initiates the currently used module"""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """
        Argument parser
        Args:
            prog_name (object): The current program
        Returns:
            object: A parser for the provided arguments
        """
        parser = super(Start, self).get_parser(prog_name)
        for argument in CONFIG:
            parser.add_argument('--'+argument, nargs='?', default='.')
        return parser

    def take_action(self, parsed_args):
        """
        Initiates the currently used module
        Args:
            Option values with the prefix "--" (str): The requested option values
        Returns:
            tuple: The execution results and stats
        """
        for argument in CONFIG:
            if getattr(parsed_args, argument) != '.':
                CONFIG[argument] = getattr(parsed_args, argument)

        module_name = CURRENT
        self.app.stdout.write('Starting ' + module_name + ':\n')
        parameters = check_parameters(CONFIG)
        if parameters != []:
            print 'The following parameters are required:\n'
            return (('Name', 'Value', 'Required'), (parameters))
        stats = manage(module_name, CONFIG)
        print '\n\nStatistics:\n'
        return (('Name', 'Value'), (stats))

class Options(Lister):
    "Shows the available options for the currently used module"

    log = logging.getLogger(__name__)

    def take_action(self, parsed_args):
        """
        Shows the available options for the currently used module
        Args:
            None
        Returns:
            tuple: Module's options
        """
        module_name = CURRENT
        self.app.stdout.write('\n\nModule '+module_name+' Options:\n')
        data = [(name, value, required, description) for name, value, required, description in META]
        for option in CONFIG:
            update_in_alist_inplace(data, option, CONFIG[option])
        return (('Name', 'Value', 'Required', 'Description'), (data))

class Error(Command):
    "Always raises an error"

    log = logging.getLogger(__name__)

    def take_action(self, parsed_args):
        self.log.info('Causing error')
        raise RuntimeError('this is the expected exception')


class LibrarySubmit(Command):
    "Submits a custom library module to library repository"

    log = logging.getLogger(__name__)

    def take_action(self, parsed_args):
        "Submits a custom library module to library repository"
        pass


class LibrarySync(Command):
    "Updates current library modules from repository"

    log = logging.getLogger(__name__)

    def take_action(self, parsed_args):
        "Updates current library modules from repository"
        pass