"""This module contains the SFADiff algorithm"""
from sfalearn.angluin_sfa import SFALearner as _SFALearner
from symautomata.cfgpda import CfgPDA as _CfgPDA
from symautomata.dfa import DFA as _DFA
from symautomata.flex2fst import Flexparser as _Flexparser
from lightbulb.core.utils.common import accept_bool, findlibrary
from lightbulb.core.utils.rcadiff import rca_diff
META = {
    'author': 'George Argyros, Ioannis Stais',
    'description': 'A black-box differential testing framework based '
                   'on Symbolic Finite Automata (SFA) learning.',
    'comments': ['SFADiff can automatically find differences between '
                 'a set of programs with comparable functionality. '
                 'Unlike existing differential testing techniques, '
                 'instead of searching for each difference individually, '
                 'SFADiff infers SFA models of the target programs '
                 'using black-box queries and systematically enumerates'
                 'the differences between the inferred SFA models']
}

class SFADiff(_SFALearner):
    """SFADiff algorithm module"""

    def __init__(self, alphabet, file, file_type, channel, cross, classic, dfa1_minus_dfa2, preload):
        super(SFADiff, self).__init__(alphabet)
        self.alphabet = alphabet
        self.membership_queries = 0
        self.cache_membership_queries = 0
        self.equivalence_queries = 0
        self.cache_equivalence_queries = 0
        self.cache_equivalence = []
        self.cache_membership = {}
        self.cross = cross
        self.channel = channel
        self.ids_membership_queries = 0
        self.ids_cache_membership_queries = 0
        self.ids_equivalence_queries = 0
        self.ids_cache_equivalence_queries = 0
        self.ids_states = 0
        self.browserstates = 0
        self.cross_check_times = 0
        self.equivalence_queries_cached_membership = 0
        self.ids_equivalence_queries_cached_membership = 0
        self.num_diff = 2
        self.dfa1_minus_dfa2 = accept_bool(dfa1_minus_dfa2)
        self.classic = accept_bool(classic)
        self.preload = accept_bool(preload)
        self.mma = None
        self.pda = None
        self.bypass = None
        self.file_type = file_type

        if file is not None and file_type is not None:
            if file_type == "FLEX":
                flex_object_m = _Flexparser(alphabet)
                self.mma = flex_object_m.yyparse(findlibrary(file))
                self.mma.minimize()
            elif file_type == "GRAMMAR":
                cfgtopda = _CfgPDA(alphabet)
                self.mma = cfgtopda.yyparse(findlibrary(file), 1)
            elif file_type == "FST":
                self.mma = _DFA()
                self.mma.load(findlibrary(file))
                self.mma.minimize()


    def getresult(self):
        """
        Returns the bypass string
        Args:
            None
        Returns:
            str: The bypass string
        """
        return  "Bypass", self.bypass

    def terminate(self):
        """
        This function should be overridden
        for SIGINT purposes in case that
        membership queries allocated resources
        """
        print 'SIGINT received. Terminating processes'
        pass


    def learn(self):
        """Initiates learning algorithm"""
        if self.preload and self.file_type != "GRAMMAR":
            return super(SFADiff, self).learn_sfa(self.mma)
        else:
            return super(SFADiff, self).learn_sfa()

    def _query(self, input_string):
        """
        The function that performs the query and should be
        overriden
        Args:
            input_string (str): Input for the target Mealy machine
        Returns:
            bool: the output of the target Mealy Machine on input input_string.
        """
        pass


    def _membership_query(self, input_string):
        """
        Consume the string input_string on the target machine (self.tmm)
        Args:
            input_string (str): Input for the target Mealy machine
        Returns:
            bool: the output of the target Mealy Machine on input input_string.
        """
        if input_string in self.cache_membership:
            self.cache_membership_queries = self.cache_membership_queries + 1
            return self.cache_membership[input_string]
        self.membership_queries = self.membership_queries + 1
        output = self._query(input_string)
        self.cache_membership[input_string] = output
        return output

    def _equivalence_query(self, mma):
        """
        Performs the equivalence query, by exchanging the learned models,
        and using the foreign model for the equivalence. If no counter
        example exists, then the algorithm falls back to the classic
        equivalence using the input automaton.
        Args:
            mma (DFA): The input automaton
        Returns:
            str: Counter example
        """

        mma = mma.concretize()
        counterexamples = []

        # Synchronization Step: IDS sends model to Browser.
        # Browser performs RCADIFF (or two-way DIFF)
        # and either terminates (diff found) or returns
        # the counterexamples (improve models)

        if self.cross == 1:

            # Only first node will count equivalence queries

            self.equivalence_queries = self.equivalence_queries + 1

            cross_mm = None

            # Browser now waits for IDS
            self.channel[0].recv()
            # receive IDS model and membership function access

            cross_mm = self.channel[3]
            crossmm = _DFA()
            crossmm.init_from_acceptor(cross_mm)
            other = self.channel[5]
            self.ids_membership_queries = self.channel[2][0]
            self.ids_cache_membership_queries = self.channel[2][1]
            self.ids_equivalence_queries = self.channel[2][2]
            self.ids_cache_equivalence_queries = self.channel[2][3]
            self.ids_equivalence_queries_cached_membership = self.channel[2][4]
            self.browserstates = len(list(mma.states))
            self.ids_states = len(list(cross_mm.states))

            # RCADIFF

            self.cross_check_times = self.cross_check_times + 1
            # ops = _RCADiff(mma, cross_mm, self.alphabet, self.left, self.right)
            # counterexamples = ops.run()

            counterexamples = rca_diff(mma, cross_mm, self.alphabet, self.num_diff, self.dfa1_minus_dfa2)

            # End of RCADIFF

            # _check for diff

            for string in counterexamples:
                if string and string != "":
                    checkbrowser = self._membership_query(string)
                    checkids = other._membership_query(string)
                    if (checkbrowser and not checkids) \
                            or (not self.dfa1_minus_dfa2 and (not checkbrowser and checkids)):
                        # A vulnerability was successfully found. Inform IDS to terminate

                        self.channel[0].send(["IDSTERMINATE"])
                        # Browser now waits for IDS
                        self.bypass = string

                        return True, None

            # No diff was found, Browser sends counterexamples (if exist) to IDS

            self.channel[0].send(counterexamples)

        if self.cross == 2:
            # IDS sends a model to Browser
            self.channel[2] = [
                self.membership_queries,
                self.cache_membership_queries,
                self.equivalence_queries,
                self.cache_equivalence_queries,
                self.equivalence_queries_cached_membership]
            self.channel[3] = mma.copy()
            self.channel[1].send(["BROWSERPROCEED"])
            # IDS now waits for Browser

            counterexamples = self.channel[1].recv()
            if counterexamples[0] == "IDSTERMINATE":
                # The IDS was instructed to terminate
                exit(1)
            counterexamples = list(reversed(counterexamples))

        # Browser and IDS examine the counterexampes and improve their model

        update = ''
        for string in counterexamples:
            if string and string != '':
                if (self._membership_query(string) and \
                                mma.consume_input(string) != True) \
                        or (not self._membership_query(string) \
                                    and mma.consume_input(string) == True):
                    update = string
                    break
        if update != '':
            self.cache_equivalence.append(update)
            return False, update

        # If no counterexample was found, and the models are equal,
        # cached data is used to self-improve the models

        # First check that our model did not learn anything wrong,
        # when making the abstraction, by testing all the cached memberships

        for string in self.cache_membership:
            if (mma.consume_input(string) != True and self._membership_query(string)) \
                    or (mma.consume_input(string) == True and not self._membership_query(string)):
                self.equivalence_queries_cached_membership += 1
                return False, string

        # Then check that the model complies with all previous cached
        # equivalence queries

        for string in self.cache_equivalence:
            if (mma.consume_input(string) != True and self._membership_query(string)) \
                    or (mma.consume_input(string) == True and not self._membership_query(string)):
                self.cache_equivalence_queries += 1
                return False, string

        # If no counterexample was found, and the models are equal, classic
        # equivalence is used to self-improve the models

        return self._classic_equivalence_query(mma)


    def _classic_equivalence_query(self, mmb):
        """
        Args:
            mmb (DFA): The equivalence
        Returns:
            tuple (bool, string): A tuple (found, ce) where found is a boolean
            value that denote whether the machine given is correct. If found is
            True then the second part of the tuple is ignored. In case found is
            False, ce should be a string in which the target machine and the
            conjectured machine give different outputs.
        """
        # mm.save('conjecture{}.dfa'.format(self.equivalenceQueries))
        if not self.classic:
            return True, None
        for string in self.cache_equivalence:
            if mmb.consume_input(string) != self.mma.consume_input(string):
                self.cache_equivalence_queries += 1
                return False, string
        self.equivalence_queries = self.equivalence_queries + 1
        mmc = self.mma.diff(mmb)
        string = mmc.shortest_string()
        if string and string not in self.cache_equivalence and self._membership_query(
                string):
            # print 'lets check '+input_string
            assert mmb.consume_input(string) != self.mma.consume_input(string), \
                "Counterexample production failed  mm(input_string):{} " \
                "self.M(input_string):{}".format(
                    mmb.consume_input(string), self.mma.consume_input(string))
            self.cache_equivalence.append(string)
            return False, string

        return True, None



    def stats(self):
        """Prints execution statistics"""
        return [
            ('Target A Membership Queries', repr(self.membership_queries)),
            ('Target A Cached Membership Queries', repr(self.cache_membership_queries)),
            ('Target A Equivalence Queries', repr(self.equivalence_queries)),
            ('Target A Cached Equivalence Queries', repr(self.cache_equivalence_queries)),
            ('Target A Cached Membership Equivalence Queries', repr(self.equivalence_queries_cached_membership)),
            ('Target B Membership Queries', repr(self.ids_membership_queries)),
            ('Target B Cached Membership Queries', repr(self.ids_cache_membership_queries)),
            ('Target B Equivalence Queries', repr(self.ids_equivalence_queries)),
            ('Target B Cached Equivalence Queries', repr(self.ids_cache_equivalence_queries)),
            ('Target B Cached Membership Equivalence Queries', repr(self.ids_equivalence_queries_cached_membership)),
            ('Learned Target B model states', repr(self.ids_states)),
            ('Learned Target A model states', repr(self.browserstates)),
            ('Cross-check times', repr(self.cross_check_times))
        ]